/**********************************************************************************
 * CHARBUF charbuf.c
 * =================
 * versione    04
 * copyright   Calzoni Pietro aka Calzo
 * licenza     Dual BSD/GPL
 * --------------------------------------------------------------------------------
 * Ora si vuole testare come funziona il timer. Per fare questo si  pensato di
 * creare un apposito file in /proc il quale riempie il buffer dei programmi che
 * lo aprono carattere x carattere, temporizzando il tutto.
 * A tale scopo vedere il cap 7 di LDD3 dove viene descritta una cosa simile e 
 * viene consigliato di non usare il comendo 'cat` il quale vuole caricare in un
 * colpo un buffer massimo di 4000 byte, ma questo dovrebbe rallentare la lettura, 
 * perch 'cat' non restituir nulla fino a che /proc/charbuf/timerstr non 
 * indicher il completamento del trasferimento dati. E` quindi + indicato usare:
 *   dd bs=1 < /proc/charbuf/timerstr
 * Per attivare i timer (e i jiffes)  necessario includere linux/timex.h
 * ================================================================================
 * BUG: /proc
 * ----
 * 1- Anche se quest versione  orientata all'uso del timer, si sono riscontrati
 * alcuni problemi con il filesystem /proc. Si noti subito che non sono stati
 * corretti nella versione 03 per poter confrontare le diversit.
 * Innanzitutto la funzione 'read_proc` definita come:
 *   int (*read_proc)
 *    ( char *buf, char **start, off_t offset, int count, int *eof, void *data )
 * Quando si vuole trasferire un (relativamente) picolo buffer, l'uso di 'buf` 
 * sufficiente, ma quando il programma che chiede di leggere richiede piccole
 * quantit di dati per volta, allora  necessario usare *start ed effettuare
 * il fill del buffer mediante *start, appunto.
 * Per maggiori info, see function 'charbuf_read_procmem_tmrstr` e 
 * 'charbuf_read_procmem_buf`
 * ================================================================================
 * PROVE: dd e cat per lettura di 72 caratteri. Il test  stato fatto su un P4 2.6G
 * single core con kernel 2.6.13.1 preentive con timer settato a HZ = 250Hz. 
 *            dd         cat
 * 1- TOD_1: 
 * real    0m7.206s    0m7.201s
 * user    0m0.000s    0m0.000s
 * sys     0m7.160s    0m7.160s
 * 2- TOD_2: (affetto dai programmi attivi)
 * real    0m7.201s    0m7.205s
 * user    0m0.000s    0m0.000s
 * sys     0m7.172s    0m7.140s
 * 3- TOD_3:
 * real    0m7.201s    0m7.200s
 * user    0m0.000s    0m0.000s
 * sys     0m0.000s    0m0.000s
 * 4- TOD_4:
 * real    0m7.202s    0m7.200s
 * user    0m0.000s    0m0.000s
 * sys     0m0.004s    0m0.000s
 *  Come si nota gli ultimi 2 metodi non danno peso a livello di CPU, mantre nei 
 * primi due casi vi  un intenso lavoro in spazio kernel. 
 *  Si osservi poi che gli ultimi due metodi sono molto pi precise dei primi: non
 * bisogna lasciarsi trarre in inganno in dal fatto che in TOD_3 e 4 dd ha dato
 * tempi diversi dai 7.2 secondi. Il tempo in pi  il tempo necesario per scrivere
 * sulla shell. Redirezionando l'output in /dev/null i valori sono identici a
 * quelli di 'cat`. Si osservi poi che il test  fatto a processore scarico. Se si
 * eseguisse un processo che richiede il 100% della CPU e si rifacesse il test, si
 * noterebbe che i tempi per i primi 2 metodi sono di circa 17 secondi, mente i
 * metodi 3 e 4 rimangono inalterati.
 * -------------------------------------------------------------------------------
 * NOTA: occorrerebbe veificare i risultati con lo stesso kernel, ma non preemptive
 * o togliendo la bigkernel look o ancora facendo variare la frequenza HZ
 *********************************************************************************/
#include <linux/init.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/kdev_t.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <asm/uaccess.h> //provide copy_from_user, copy_to_user, access_ok
#include "charbuf.h"
//--- /proc ---
#include <linux/proc_fs.h>
//--- ioctl ---
#include <linux/ioctl.h> //Documentation/ioctl-number.txt x avere qualche "numero magico"
#include "cb_ioctl.h"
//--- temporizzazioni ---
#include <linux/jiffies.h> //andrebbe bene anche linux/sched.h
#include <linux/timex.h>   //non so se serva veramete

MODULE_LICENSE("Dual BSD/GPL");
MODULE_VERSION("0.4");

/** Parametri del modulo. E' fondamentalmente la dimensione dell'array. Nota che
 * S_IRUGO indica che ciunque pu leggere il parametro
 * S_IWUSR solo root pu modificare il parametro (tramite sysfs se non erro)
 * altre definizioni si possono trovare in linux/stat.h
 * */
static unsigned long bufdimension  = DEF_BUF_DIMENSION;
module_param(bufdimension, ulong, S_IRUGO|S_IWUSR); // see LDD3 cap 2, pag 36
MODULE_PARM_DESC(bufdimension, "User paramether of initial buffer dimension");

struct charbuf_dev *charbuf_device; // solo un device

//Strutture relative al /proc filesystem
struct proc_dir_entry *proc_charbufdir=NULL;
struct proc_dir_entry *proc_charbufdim=NULL; //permette l'accesso al file virtuale charbufdim
struct proc_dir_entry *proc_charbuffer=NULL;
struct proc_dir_entry *proc_chartmrstr=NULL;
#define PROC_CHARBUFDIR "charbuf"   //subdir per /proc
#define PROC_CHARBUFDIM "dimension" //file virtuale per ottenere dimensione e index
#define PROC_CHARBUFFER "buffer"    //file virtulae per accedere al buffer
#define PROC_TIMERSTR   "timerstr"  //file virtuale ad "accesso temporizzato"

// --- timer ---
/** 4 tipi di timer: per settarli scrivere 0, 1, 2, 3 o 4 nel file /proc/charbuf/timerstr */
#define TOD_0 0 // 0: nessun timer e la sctringa viene passata istantaneamente
#define TOD_1 1 // 1: utilizzo di cpu_relax()
#define TOD_2 2 // 2: utilizzo di schedule()
#define TOD_3 3 // 3: utilizzo di wait_event_interruptible_timeout
#define TOD_4 4 // 4: utilizzo di schedule_timeout e relativo set di modalit del task (run, interruptible, ecc)
/** type_of_delay definisce il tipo di delay da usare nella funzione 
 * 'charbuf_read_procmem_tmrstr` 
 * */
static int type_of_delay = TOD_4;
/** Volendo realizzare 0.5 secondi di attesa a carattere, ococorre basarsi sulla
 * configuraizione del timer nel kernel mediante la macro HZ
 * */
#define JIFFIES_ONE  (HZ)      //   1 secondo
#define JIFFIES_HALF (HZ >> 1) // 0.5 secondi
#define JIFFIES_DEC  (HZ / 10) // 0.1 secondi


/** La struttura seguente definisce le funzioni di accesso al file */
struct file_operations charbuf_fops =
{
   //serve anche per a comunicazione a nel ringbuffer del kernel credo
   .owner = THIS_MODULE,
   //Open
   .open  = charbuf_open,
   //Scrive
   .write = charbuf_write,
   //Legge
   .read  = charbuf_read,
   //IO-Ctl
   .ioctl = charbuf_ioctl,
};


/**********************************************
 * Funzioni per IO-Ctl
 **********************************************/
/** _ioctl permette di accedere al device in modo indiretto. Per prima cosa occorre
 * controllare che il magic number sia quello giusto e di non superare il numero di
 * comandi ammessi, altrimenti return -ENOTTY (ossia "inappropriate IOCtl)
 * */
int charbuf_ioctl(struct inode *inode, struct file *filp,
		  unsigned int cmd, unsigned long arg)
{
   struct charbuf_dev *dev = filp->private_data;
   int err=0, retval=0;
   int i,iaux; //integer ausiliary value
   char *auxbuf;
   
   printk(KERN_ALERT "Richiesto accesso a ioctl func (%i)...\n", cmd);
   if(_IOC_TYPE(cmd) != CB_IOC_MAGIC) //controllo che sia il comando richiesto
     return -ENOTTY; 
   if(_IOC_NR(cmd) > CB_IOCTL_MAXNR) //se eccedo con i comandi non va bene. Ma non dovrebbe essere ">=" ?
     return -ENOTTY;
   
   if(_IOC_DIR(cmd) & _IOC_READ) //Le 4 righe che seguono sarebbero da riscrivere meglio perch cos non hanno molto senso...
     err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
   if(_IOC_DIR(cmd) & _IOC_WRITE) //... questo if mi elimina l'operazione precedente. Insensato, ma fa lo stesso per ora
     err = !access_ok(VERIFY_WRITE, (void __user *)arg, _IOC_SIZE(cmd));
   if (err) 
     return -EFAULT;
   
   //Processing dei comandi:
   switch(cmd)
     {
      case CB_IOCTL_QSIZE:
	return dev->size;
	break; //inutile a causa del return
	
      case CB_IOCTL_QINDEX:
	return dev->index;
	break;
	
      case CB_IOCTL_SSIZE: //setto il nuovo valore
	if(down_interruptible(&dev->sem))
	  return -EBUSY; //preferiso BUSY a RESTARTSYS perch se siamo qui secondo me  facile che il perch sia legato al fatto che qualcun'altro ha accesso al device
	
	retval = __get_user(iaux, (int __user *)arg); //S indica che leggo da un buffer!
	if(iaux > MAX_BUF_DIMENSION)
	  iaux = MAX_BUF_DIMENSION;
	if(iaux != dev->size) //se sono uguali non fare nulla 
	  {
	     if(!(auxbuf = kmalloc(iaux, GFP_KERNEL)))
	       {	
		  printk(KERN_ALERT "Not enought memory.\n");
		  retval = -ENOMEM;
	       } else
	       {
		  if(iaux < dev->index)
		    dev->index = iaux;
		  printk(KERN_ALERT "Redimensioning buffer to %i byte throught IO-Ctl\n", iaux);
		  for(i=0; i<dev->index; i++)
		    auxbuf[i]=dev->charbuffer[i];
		  kfree(dev->charbuffer);
		  dev->charbuffer=auxbuf;
	       }
	     dev->size = iaux;
	  }
			
	up(&dev->sem);
	break;
	
      case CB_IOCTL_SBUF: //setta il contenuto del buffer
	if(down_interruptible(&dev->sem))
	  return -EBUSY;
	
	struct cb_buf *usbuf; //userspace buffer
	retval = __get_user(usbuf, (struct cb_buf __user **)arg);
	if(capable(CAP_SYS_ADMIN)) //se sei amministratore puoi scrivere oltre .index
	  {
	     if(usbuf->dim > dev->size) //non puoi scrivere oltre .size, per!
	       {
		  usbuf->dim = dev->size;
		  dev->index = dev->size;
	       }
	     else
	       dev->index = usbuf->dim;
	  }
	else if(usbuf->dim > dev->index)
	  {
	     usbuf->dim = dev->index;
	     printk(KERN_ALERT "Only root can write more then .index\n");
	  }
	else 
	  dev->index = usbuf->dim;
	
	for(i=0; i<usbuf->dim; i++)
	  dev->charbuffer[i] = usbuf->buf[i];
	
	up(&dev->sem);
	break;
     }
   
   return retval;
}



/**********************************************
 * Funzioni generiche di I/O (POSIX)
 **********************************************/

/** _open non fa altro che resettare il buffer se il device driver  aperto in
 * scrittura.
 * NOTA: per verificare solo questa procedura  sufficiente eseguire da shell:
 *   touch nome_device
 * oppure
 *   <, >, >> nome_device (ossia leggi, apri resettando, apri aggiungendo)
 * */
int charbuf_open(struct inode *inode,
		 struct file  *flip)
{
   struct charbuf_dev *dev;

   printk(KERN_ALERT "Open charbuffer\n");
   dev = container_of(inode->i_cdev, struct charbuf_dev, cdev);
   if(!dev) 
     printk(KERN_ALERT "WARNING: charbuffer device is null!\n");
   //il kernel preleva questo dato e lo mette nella variabile file che servir poi a write e read
   flip->private_data = dev; // qui credo che in questo caro si potrebbe mettere `charbuf_device'

   if((flip->f_flags & O_APPEND) == O_APPEND) //APPEND
     {
	printk(KERN_ALERT "Append.\n"); //segnalo la cosa, ma non faccio nulla
     }
   else
   if((flip->f_flags & O_ACCMODE) == O_WRONLY)
    {
       if(down_interruptible(&dev->sem))
	 return -ERESTARTSYS;
       printk(KERN_ALERT " Open: reset buffer.\n");
       memset(dev->charbuffer,0, dev->size);
       dev->index=0;      
    }
   
   up(&dev->sem);
      
   return 0;
}


/** _write: scrive nel buffer a partire da una determinata posizione. Chiaramente
 * deve prevedere un controllo in modo da non eccedere il buffer.
 * */
ssize_t charbuf_write(struct file *flip,
		      const char __user *buf, //__user: buf in user space = ATTENTO A QUELLO CHE FAI
		      size_t count,
		      loff_t *f_pos)
{
   //quanto segue si pu fare grazie alle prime righe di _open!!
   struct charbuf_dev *dev = flip->private_data; //con questo "estrapolo" la struttura del dispositivo, in altre parole la variabile globale charbuf_device in questo caso
   ssize_t returnvalue=0;
   int err;
   
   printk(KERN_ALERT "Write\n count=%i\n",count);
   if(down_interruptible(&dev->sem)) //valuto se il dispositivo  in uso (credo)
     return -ERESTARTSYS;
   
   if(*f_pos >= dev->size) //il buffer  unico e se la posizione del file in cui voglio scrivere  maggiore o =,  errato
     {
	returnvalue=-ENOSPC;
	goto exit_write_fcn; //esci dalla funzione write
     }

   //Esamino il tipo di apertura
   if((flip->f_flags & O_APPEND) == O_APPEND) //APPEND
     {
	//se sono gia alla fine del buffer non ha senso proseguire...
	if(dev->index >= dev->size-1)
	  {
	     returnvalue=-ENOSPC;
	     goto exit_write_fcn; //esci dalla funzione write
	  }
	else
	  if(*f_pos + count + dev->index >= dev->size) //se supero la dimensione
	    count = dev->size - (*f_pos + dev->index);
	err = copy_from_user(&dev->charbuffer[*f_pos + dev->index], buf, count);
     }
   else //se non  un append allora  un write only (quindi un rewrite)... il reset lo faccio con la open function
     {
	if(*f_pos + count > dev->size)
	  count = dev->size - *f_pos;
	err = copy_from_user(&dev->charbuffer[*f_pos], buf, count);
     }
 
   if(err)
     {
	returnvalue = -EFAULT;
	goto exit_write_fcn;
     }
   
   *f_pos += count; //aggiorno il file pointer
   returnvalue = count; //restituisco il numero di byte copiati
   dev->index += count; //Aggiorno l'index
   
exit_write_fcn:
   up(&dev->sem);
   return returnvalue;
}
  
/** _read: legge il buffer
 * */
ssize_t charbuf_read(struct file *flip,
		     char __user *buf,
		     size_t count,
		     loff_t *f_pos)
{
   struct charbuf_dev *dev = flip->private_data; //possibile grazie alle prime righe di _open
   ssize_t returnval=0;
   
   printk(KERN_ALERT "Read.\n");
   if(down_interruptible(&dev->sem))
     return -ERESTARTSYS;

   printk(KERN_ALERT "size = %i\nidx  = %i\nf_pos= %i\ncount= %i\n",
	  (int)dev->size, dev->index, (int)*f_pos, (int)count);

   if(*f_pos >= dev->size) //se la posizione leneare  superiore alla dimensione am$
     count = 0; //in teoria potrei fare semplicemente 'return 0' ma preferisco fare cos per avere un feedback di debug
   else
   if(*f_pos + count >= dev->index) //attenzione! qui  index e non size altrimenti un cat del device d un loop infinito... e a pensarci (molto) bene  giusto che sia cos
     count = dev->index - *f_pos;
   
   printk(KERN_ALERT "copy %i byte...\n", count);
   if(copy_to_user(buf,dev->charbuffer + *f_pos, count))
     {
	printk(KERN_ALERT "Error while copy to user space!\n");
	returnval = -EFAULT;
	goto exit_read_fnc;
     }
   
   exit_read_fnc:
   *f_pos += count;
   returnval=count;
   up(&dev->sem);
   
   return returnval; //se 0  perch nulla  copiato e quindi blocco la lettura
}


  
/**********************************************
 * Funzioni di gestione
 **********************************************/

/** _menage_buffer serve per maneggiare il buffer. Realloca il buffer senza 
 * perderne il contenuto nell'eventualit in cui il buffer riallocaro sia pi
 * grande di prima. Chairamente deve essere aggiornato anche il campo 'size'
 * @parm dev  il puntatore a charbuf_dev ossia il dispositivo vero e proprio che
 * contiene il puntatore al buffer da riallocare.
 * @parm memoria da rialocare.
 * @return 0 se tutto bene, <0 errore di allocazione, >0 ok, ma dati troncati 
 * ATTUALMENTE NON SVILUPPATA*/
int charbuf_menage_buffer(struct charbuf_dev *dev,int new_dimension)
{
   //TODO
   return 0;
}

/** _init_buffer alloca solo la memoria. Viene chiamata (in teoria) solo da
 * _init_module.
 * @parm dev  il puntatore a charbuf_dev ossia il dispositivo vero e proprio che
 * contiene il puntatore al buffer da riallocare.
 * @parm memoria da allocare
 * @return 0 se tutto bene, <0 errore di allocazione
 * */
int charbuf_init_buffer(struct charbuf_dev *dev,int dimension)
{
   if(!dev)
     {
	printk(KERN_ALERT "Error: why charbuf_dev is NULL?\n");
	return -ENOMEM;
     }
   
   if(dimension>MAX_BUF_DIMENSION)
     {
	printk(KERN_ALERT "Warning: dimension exceed max dimension %i byte\n",
	       MAX_BUF_DIMENSION);
	dev->size=MAX_BUF_DIMENSION;
     }
   else
     dev->size=dimension;
   
   if(!(dev->charbuffer=kmalloc(dev->size, GFP_KERNEL)))
     {
	printk(KERN_ALERT "Error: not enougth memory\n");
	return -ENOMEM;
     }
   
   return 0;
}


/**********************************************
 * Funzioni di accesso al filesystem /proc    *
 **********************************************/
/** _read_procmem legge/crea dei dati da passare in fase di accesso ai files
 * contenuti in /proc. Credo che esista una funzione di read per ogni file a cui
 * si richiede l'accesso al file PROC_CHARBUFDIM.
 * Un esempio di funzione  a pagina 85 di LDD3 ('implement file in /proc'). Lo
 * stesso LDD3 dice che l'esempio  un po' brutto e credo si riferisca la fatto che
 * il sistema non presenta un controllo relativo al buffer che st scrivendo;
 * buffer che comunque  in user-space quindi non dovrebbe causare problemi di
 * sicurezza, ma non ne sono troppo sicuro.
 * Io quindi ho cercato di migliorare (se possibile) la cosa andando prima a creare
 * la stringa, e quindi andando a copiarla nel buffer a poco a poco (byte x byte)
 * in modo da tenere sotto controllo il limite (count) del buffer.
 * Altra cosa molto importante  la gestione dell'offset che viene fatta nella fase
 * iniziale. Ci rende la procedura pi completa nell'eventualit che venga letto
 * il file a partire da un punto maggiore del primo byte (offset>0).
 * */
int charbuf_read_procmem_dim(char *buf, char **start, off_t offset,
			     int count, int *eof, void *data)
{
   int numlen=0, len=0, i; //per precauzione pongo la lunghezza pari all'offset
   char num[12]; //dovrebbero bastare 11 byte: "4294967295"+'\n'
   struct charbuf_dev *device=charbuf_device; //setto il device

   if(down_interruptible(&device->sem))
     return -ERESTARTSYS;

   if(!buf) //se il buffer  nullo  meglio uscire ;) ... anche se non dovrebbe mai verificarsi
     goto exit_func;
   
   numlen=sprintf(num,"%u\n",device->size);
   printk("Proc_Read_DIM o=%u  c=%u  n=%u\n",offset,count,numlen);
   if(offset < numlen) //creo la stringa e verifico di non scrivere troppo ;)
     i=offset;
   else
     {  
	offset-=numlen; //decremento offset perch
	numlen=sprintf(num,"%u\n",device->index);
	if(offset < numlen)
	{
	   i=offset;
	   goto next_value;
	}
	else
	  goto exit_func;
     }
   
   do //trasferisco il primo numero
     {
	if(len<count) // devo tenere sotto controllo anche la massima lunghezza
	{
	   buf[len++]=num[i];
	   printk(KERN_INFO "%c\n",num[i]);
	} else
	  goto exit_func;
     }
   while(num[i++]!='\n'); //questo while mi permette di scrivere tutto fino all'ultimo carattere '\n'
   
   numlen=sprintf(num,"%u\n",device->index); //creo la seconda stringa
   i=0;
   next_value:
   
   do //trasferisco il secondo numero
     {
	if(len<count) // devo tenere sotto controllo anche la massima lunghezza
	  buf[len++]=num[i];
	else
	  goto exit_func;
     }
   while(num[i++]!='\n');
   
   exit_func: //prima di *eof=1 perch altrimenti ho paura che le unzioni un user-space non si accorgano che il file  finito
   *eof=1;
   up(&device->sem);
   
   return len;
}

/** _read_procmem legge anche in questo caso la memoria, ma adesso si parla del
 * buffer vero e proprio. Vale ovviamente tutto ci che  stato detto per la 
 * funzione `charbuf_read_procmem_dim'.
 * Si noti che in questa funzione, a differenza della precedente, alla fine viene 
 * notificato l'end-of-file a seguito di un 'if'. Questo rende migliore il
 * comportamento del device a seguito della lettura del buffer tramite /proc: si
 * riesce infatti a capire si siamo alla fine del file oppure no.
 * ------------------------------------------------------------------------------
 * La funzione deve prevedere l'accesso a programmi come `dd' i quali possono
 * cominciare a leggere i file non dall'inizio, ma da un qualsiasi punto, ed
 * inoltre pu richiedere che il buffer in user-space sia caricato con blocchi
 * di diversa dimensione (count).
 * Per fare ci occorre come noto indicare al programma quando  raggiunto l'eof,
 * ma anche restituire il numero di byte trasferiti, ossia 'i` (e non len come
 * nella versione 03 precedente).
 * Si noti che a differenza delle _read normali, in questo caso  sufficiente 
 * *eof=1 per indicare il terminamento del processo di fill del buffere 'buf`,
 * mentre nelle _read normali il processo di lettura si arresta a fronte di un
 * return 0.
 * Ultima nota fondamentale  che per trasferire un butter di dimensione X tramite
 * n pacchetti pi piccoli di dimensione Y, occorre aggiungere *start=buf che, se
 * non ho cpito male da LDD3 cap4 pag 84, permette di usare una sorta di 
 * "paginazione".
 * */
int charbuf_read_procmem_buf(char *buf, char **start, off_t offset,
			     int count, int *eof, void *data)
{
   // len, ossia offset, indica da che punto del buffer sorgente si comincia a leggere
   // i  un contatore, ma anche il ritorno della funzione che indica quanti byte sono stati copiati
   int len=offset, i;
   struct charbuf_dev *device = charbuf_device;
   
   if(down_interruptible(&device->sem))
     return -ERESTARTSYS;
   
   if(offset >= device->index)
     { 
	*eof = 1; //segnalo che la parte utile del buffer  esaurita
	i = len;  //anche se non  vero, devo passare questo valore di byte copiati per una correta esecuzione di dd
	goto exit_func;
     }
   
   if(count + offset >= device->index)
     count = device->index - offset; //qui offset  sicuramente < di .index
   
   *start=buf; //necessatio per procedere alla lettura del buffer copiando piccole quantit di dati per volta
   for(i=0; i<count; i++) //fase di copiatura dei singoli byte
     *(*start + i) = device->charbuffer[ len++ ]; //buf[i]=device->charbuffer[len++]; questa  la linea versione 03
     
   exit_func:

   up(&device->sem);
   // nella versione 03 la riga sotto era 'return len;'
   return i; //restituisco quanti byte sono stati copiati
}

/** _read_procmem legge il file e stampa una determinata stringa, ma la passa allo
 * user-space carattere x carattere temporizzando il processo. Questa funzione
 *  stata pensata per accedere al file mediante il comando
 *   dd if=/proc/charbuf/timerstr bs=1
 * dove bs  un numero a caso minore della dimensione del file virtuale 'timerstr`
 * --------------------------------------------------------------------------------
 * NOTA: sizeof(CHARBUF_PROC_STR)  la lunghezza della stringa + il carattere '\0'
 * di terminazione, quindi quando si gestiscono gli indici devo sempre sottrarre 1
 * a questo valore per non rischiare di legere aree di memoria protetta
 * --------------------------------------------------------------------------------
 * Per testare questa funzione occorre, si puossono eseguire i seguenti comandi:
 *   time cat /proc/charbuf/timerstr
 *   time dd if=/proc/charbuf/timerstr bs=1
 * nel primo caso dopo un tempo pari a   sizeof(CHARBUF_PROC_STR)*JIFFIES_DEC
 * compare la stringa intera. Nel secondo caso compariranno i caratteri a
 * intervalli di JIFFIES_DEC.
 * Per settare il tipo di temporizzazione, scrivere in timerstr il numero corretto.
 * Nota: JIFFIES_DEC  tale da scrivere un carattere ogni decimo di secondo, quindi
 * se la stringa  di 72 caratteri, il tempo per scriverla  7,2 secondi.
 * */
#define CHARBUF_PROC_STR "CharBuf: modulo indipendente dall'architettura\n\t-- \033[31mby Calzo\033[0m --\n"
int charbuf_read_procmem_tmrstr(char __user *buf, char **start, off_t offset,
				int count, int *eof, void *data)
{
   int len=offset, i=0; //len=0 chiude la procedura di lettura
   const char bufstr[] = CHARBUF_PROC_STR;
   int j1;
   wait_queue_head_t wait;
   
   if (offset >= (sizeof(CHARBUF_PROC_STR))-1) //qesto if  precauzionale
     {
	*eof = 1; //Nei file virtuali occorre sempre segnalare l'eof al posto che restituire un numero nullo di byte copiati
	i = len;  //FONDAMENTALE!! se non lo metto dd non funzione bene. NOTA: len=offset
	goto exit_func;
     }
   
   if ( count+offset >= (sizeof(CHARBUF_PROC_STR))-1 )
     count = sizeof(CHARBUF_PROC_STR)-offset-1;
   
   *start=buf; //start serve perc se scrivo "+ pagine" non posso usare buf, ma devo assegnare proprio start. Questo si vede se uso x es: 'dd bs=2 if=...'
   
   switch ( type_of_delay )
     {
	/** TOD_0 non fa nulla: trasferisce la stringa istantaneamente */
      case TOD_0: // nothing  
	for (i=0; i < count; i++)
	  *(*start+i) = *(bufstr+len+i);
	break;
	
	/** TOD_1 usa cpu_relax() che ha un problema: se il kernel non  reattivo
	 * (preemptive) il while loop che segue tende a confrontare continuamente
	 * i jiffies usando pesantemente la CPU */
      case TOD_1: // delay with cpu_relax() (che di relax ha poco visto che il processore va al 100%
	for (i=0; i < count; i++)
	  {
	     j1=jiffies + JIFFIES_DEC; //primaleggo i jiffies poi eseguo le operazioni
	      *(*start+i) = *(bufstr+len+i);
	     while (time_before(jiffies, j1))
	       cpu_relax(); // funzione strettamente legata all'architettura
	  }
	break;
	
	/** TOD_2 usa 'schedule` che  migliore della precedente, ma se non vi
	 * sono altri processi in coda, non apporta miglioramenti in termini di
	 * prestazioni perch lo scheduler tender ad eseguire sempre il processo
	 * corrente se nessun'altro richide l'uso della CPU */
      case TOD_2:
	for (i=0; i < count; i++)
	  {
	     j1=jiffies + JIFFIES_DEC; //primaleggo i jiffies poi eseguo le operazioni
	     *(*start+i) = *(bufstr+len+i);
	     while (time_before(jiffies, j1))
	       schedule();
	  }
	break;
	
	/** TOD_3 usa 'wait_event_interruptible_timeout` che migliora nettamente le
	 * prestazioni: infatti non occupa CPU inutilamente. Funzioni correlate:
	 * 'wait_event_interruptible`, 'wait_event`, 'wait_timeout` LDD3 pag 149 
	 * Se questa funzione restituisce un numero negativo  perch il carico di
	 * CPU  tale per cui ha ritardato l'esecuzione di questo task */
      case TOD_3:
	for (i=0; i < count; i++)
	  {
	     init_waitqueue_head (&wait); // inizializzo la "coda", ossia determino l'istante iniziale
	     *(*start+i) = *(bufstr+len+i);
	     wait_event_interruptible_timeout(wait, 0, JIFFIES_DEC); // aspetto
	  }
	break;
	
	/** TOD_4 usa 'schedule_timeout` che  forse ancora meglio, perch oltre a
	 * chiamare questa funzione, si pu definire come si deve comportare il
	 * task ossia se per esempio  interruttibile o no */
      case TOD_4:
	for (i=0; i < count; i++)
	  { //alcune opzioni (de sched.h): TASK_INTERRUPTIBLE, TASK_RUNING, TASK_UNINTERRUPTIBLE
	     set_current_state(TASK_INTERRUPTIBLE);
	     *(*start+i) = *(bufstr+len+i);
	     schedule_timeout (JIFFIES_DEC);
	     set_current_state(TASK_RUNNING); // questo probabilamente non serve
	  }
	break;
	
      default:
	printk(KERN_ALERT "Warning: invalid delay option!\n");
	break;
     }

   exit_func:
   
//   printk(KERN_ALERT "proc_tmrstr: off=%i, count=%i i=%i\n",
//	  (unsigned)offset, (unsigned)count, i);
   
   return i; //devo restituire count, ossia il numero di byte copiati
}

/** _write_procmem questa prima write permette di effettuare la scrittura del file
 * dimension. Scrivendo 'x' seguito da un solo numero, moltiplico la dimensione
 * del buffer (entro i limiti), mentre se al posto di 'x' vi  ':', allora si
 * divider il valore.
 * */
int charbuf_write_procmem_dim(struct file *file, const char __user *buf,
			      unsigned long count, void *data)
{
   int i, newbufdim=0;
   char b[2], *auxbuf=NULL;
   struct charbuf_dev *device=charbuf_device;
   
   if(down_interruptible(&device->sem))
     return -ERESTARTSYS;
     
   if(!buf) //probabilmente superfluo
     goto exit_func;
   
   if(count>=2)
     copy_from_user(b, buf, 2); //prendo solo i primi 2 byte.
   else
     goto exit_func;
   
   if((newbufdim=b[1]-'0')>9) //massimo 9...
     goto exit_func;
   else if(newbufdim<2)//... e minimo 2 perch 0 non pu essere (il buf deve esistere) e 1 copierebbe operazioni inutili
     goto exit_func;
   
   auxbuf=device->charbuffer; //"sposto" il buffer su un altro puntatore ausiliario
   if(b[0]=='x') //fase di moltiplicazione
     {
	device->size*=newbufdim;
	if(device->size>MAX_BUF_DIMENSION)
	  device->size=MAX_BUF_DIMENSION;
	if(!(device->charbuffer=(char*)kmalloc(device->size,GFP_KERNEL)))
	  {
	     printk(KERN_ALERT "Not enought memory!\n");
	     goto exit_func;
	  }
	for(i=0; i<device->index; i++)
	  device->charbuffer[i]=auxbuf[i];
     } 
   else if(b[0]==':') //fase di divisione
     {
	device->size/=newbufdim; //minimo devo mettere 1 byte
	if(device->size<=0) 
	  device->size=1; //minimo 1 byte
	if(!(device->charbuffer=(char*)kmalloc(device->size,GFP_KERNEL)))
	  {
	     printk(KERN_ALERT "Not enought memory!\n");
	     goto exit_func;
	  }	
	for(i=0; i<device->size && i<device->index; i++)
	  device->charbuffer[i]=auxbuf[i];
	device->index=i;
     }

   printk(KERN_ALERT "Redimensioning buffer to %i byte through /proc\n", device->size);
   exit_func:
   if(!auxbuf) kfree(auxbuf);
   up(&device->sem);
   return count; //restituisco lo stesso valore di count iniziale in modo da terminare sempre l'invio di dati
}

/** _write_procmem: se viene scritto 0 con
 *    echo 0 > /proc/charbuf/buffer
 * il buffer si annulla (memeset)
 * */
int charbuf_write_procmem_buf(struct file *file, const char __user *buffer,
			      unsigned long count, void *data)
{
   struct charbuf_dev *device = charbuf_device;
   char c;
   
   if(!count) //se  0 non fare nulla
     return count;
   
   if(down_interruptible(&device->sem))
     return -ERESTARTSYS;
   
   copy_from_user(&c, buffer, 1);
   if(c=='0') //annullo il buffer
     {
	memset(device->charbuffer, 0, device->size);
	device->index=0;
	printk("Clean buffer (%i byte)\n", device->size);
     }
   
   up(&device->sem);
   return count; //return count mi permette di fare tutto in una sessione
}

/** _write_proc: questa write permette di settare il tipo di delay (TOD) scrivendo
 * 0, 1, 2, 3 o 4 nel file virtuale timerstr
 * */
int charbuf_write_procmem_tmrstr(struct file *file, const char __user *buffer,
				 unsigned long count, void *data)
{
   struct charbuf_dev *device = charbuf_device;
   char c;
   
   if(!count) //se  0 non fare nulla
     return count;
   
   if(down_interruptible(&device->sem))
     return -ERESTARTSYS;
   
   copy_from_user(&c, buffer, 1);
   if(( c >= '0' ) && ( c <= '4' ))
     {
	type_of_delay = c - '0';
	printk(KERN_ALERT "Set type of delay: %d\n", type_of_delay);
     }
   
   
   up(&device->sem);
   return count; //return count mi permette di fare tutto in una sessione
}



/**********************************************
 * Inizializzazione e deallocamento del modulo
 **********************************************/
/** La funzione di setup credo che serva per indicare al kernel quali sono le 
 * funzioni di I/O (posix credo). Questa funzione viene chiamata alla fine di 
 * _init_module() passandogli il device "charbuf_device" (quindi dev=charbuf_device)
 * e index in questo caso non dovrebbe servire perch il device che verr creato
 *  solo uno.
 * */
void charbuf_setup_cdev(struct charbuf_dev *dev, int index)
{
   int err, dev_number=MKDEV(CHARBUF_MAJOR,0+index);
   
   cdev_init(&dev->cdev, &charbuf_fops);
   //qui ribadisco al kernal quali sono le funzioni posix "compilando" la struct cdev
   dev->cdev.owner = THIS_MODULE;
   dev->cdev.ops = &charbuf_fops;
   err = cdev_add(&dev->cdev, dev_number, 1); //0 o 1?
   if(err != 0)
     printk(KERN_NOTICE "Error %d adding charbuf%d", err, index);
}


/** _init_module non fa altro che inizializzare il buffer nel kernel.
 * */
int charbuf_init_module(void)
{
   dev_t dev;
   
   dev = MKDEV(CHARBUF_MAJOR,CHARBUF_MINOR);
   printk(KERN_INFO "Initailization of %s (major-minor %i-%i)\n",
	  DEVICE_NAME,CHARBUF_MAJOR,CHARBUF_MINOR);
   //Registrazione
   if(register_chrdev_region(dev,1,DEVICE_NAME)<0) //0 o 1?
     printk(KERN_ALERT "Error while register %s!\n",DEVICE_NAME);
   else
     printk(KERN_ALERT "Registration done.\n");
   
   //A questo punto alloco il device... 
   charbuf_device=kmalloc(sizeof(struct charbuf_dev), GFP_KERNEL);
   if(!charbuf_device)
     {
	printk(KERN_ALERT "Fatal: unable to alloc memory for %s device\n", DEVICE_NAME);
	return -ENOMEM;
     }
   memset(charbuf_device,0,sizeof(struct charbuf_dev));
   //...e anche il buffer
   printk(KERN_ALERT "Alloc memory (%i byte)\n",(int)bufdimension);
   if(charbuf_init_buffer(charbuf_device,bufdimension)<0)
     {
	kfree(charbuf_device);
	return -ENOMEM;
     }
   
   //init dei semafori
   init_MUTEX(&charbuf_device->sem);
   charbuf_setup_cdev(charbuf_device,0);
   
   //Versione 02: intorduzione del filesystem /proc. con "create_proc_read_entry"
   //creo e inizializzo la struttura
   proc_charbufdir=proc_mkdir(PROC_CHARBUFDIR,NULL); //creo la subdir in proc
   if(!proc_charbufdir)
     {
	printk(KERN_ALERT "WARNING: proc_charbuf_dir NULL!\n");
     } 
   else
     {
	proc_charbufdim = create_proc_read_entry(PROC_CHARBUFDIM, /*virtual filename*/ 
						 0, /*default mode*/
						 proc_charbufdir, /*parent dir*/
						 charbuf_read_procmem_dim, /*funzione scritta da me*/
						 NULL /*data client... non ho ben capito a cosa serve*/
						 );
	proc_charbuffer = create_proc_read_entry(PROC_CHARBUFFER, 
						 0, proc_charbufdir,
						 charbuf_read_procmem_buf, NULL);
	proc_chartmrstr = create_proc_read_entry(PROC_TIMERSTR,
						 0, proc_charbufdir,
						 charbuf_read_procmem_tmrstr, NULL);
	if(!proc_charbufdim || !proc_charbuffer || !proc_chartmrstr)
	  {
	     printk(KERN_ALERT "WARNING: wrong init /proc filesystem!\n");
	  }
	else
	  {
	     proc_charbufdim->write_proc=charbuf_write_procmem_dim;
	     proc_charbuffer->write_proc=charbuf_write_procmem_buf;
	     proc_chartmrstr->write_proc=charbuf_write_procmem_tmrstr;
	  }
     }

   return 0;
}

/** _cleanup_module libero la memoria
 * NOTA: charbuf_device->cdev.ops=NULL;  fondamentale per evitare segfault o
 * congelamenti dovuti alla rimozione del driver ma alla non cancellazione del
 * file di dispositivo.
 * */
void charbuf_cleanup_module(void)
{
   dev_t dev_number=MKDEV(CHARBUF_MAJOR, 0);
   
   charbuf_device->cdev.ops=NULL;
   cdev_del(&charbuf_device->cdev);
   
   if(charbuf_device->charbuffer)
     kfree(charbuf_device->charbuffer);
   if(charbuf_device)
     kfree(charbuf_device);
   charbuf_device=NULL;

   //Versione 02: Aggiungo l'eliminazione del proc entry file
   remove_proc_entry(PROC_TIMERSTR,   //file name
		     proc_charbufdir //parent dir
		     );
   remove_proc_entry(PROC_CHARBUFDIM, //file name
		     proc_charbufdir //parent dir
		     ); 
   remove_proc_entry(PROC_CHARBUFFER, //file name
		     proc_charbufdir //parent dir
		     );
   remove_proc_entry(PROC_CHARBUFDIR, //file name
		     NULL //parent dir, null=/proc
		     );
   
   printk(KERN_ALERT "Unregister char region (%s)\n",DEVICE_NAME);
   unregister_chrdev_region(dev_number, 1); //0 o 1?
}

module_init(charbuf_init_module);
module_exit(charbuf_cleanup_module);
